@@ -1,5 +1,5 @@
 /* libmspack -- a library for working with Microsoft compression formats.
- * (C) 2003-2004 Stuart Caie <kyzer@4u.net>
+ * (C) 2003-2011 Stuart Caie <kyzer@4u.net>
  *
  * libmspack is free software; you can redistribute it and/or modify it under
  * the terms of the GNU Lesser General Public License (LGPL) version 2.1
@@ -113,6 +113,41 @@
  * - #MSPACK_ERR_CHECKSUM indicates that a data checksum has failed.
  * - #MSPACK_ERR_CRUNCH indicates an error occured during compression.
  * - #MSPACK_ERR_DECRUNCH indicates an error occured during decompression.
+ *
+ * \section threading Multi-threading
+ *
+ * libmspack methods are reentrant and multithreading-safe when each
+ * thread has its own compressor or decompressor.
+
+ * You should not call multiple methods simultaneously on a single
+ * compressor or decompressor instance.
+ *
+ * If this may happen, you can either use one compressor or
+ * decompressor per thread, or you can use your preferred lock,
+ * semaphore or mutex library to ensure no more than one method on a
+ * compressor/decompressor is called simultaneously. libmspack will
+ * not do this locking for you.
+ *
+ * Example of incorrect behaviour:
+ * - thread 1 calls mspack_create_cab_decompressor()
+ * - thread 1 calls open()
+ * - thread 1 calls extract() for one file
+ * - thread 2 simultaneously calls extract() for another file
+ *
+ * Correct behaviour:
+ * - thread 1 calls mspack_create_cab_decompressor()
+ * - thread 2 calls mspack_create_cab_decompressor()
+ * - thread 1 calls its own open() / extract()
+ * - thread 2 simultaneously calls its own open() / extract()
+ *
+ * Also correct behaviour:
+ * - thread 1 calls mspack_create_cab_decompressor()
+ * - thread 1 locks a mutex for with the decompressor before
+ *   calling any methods on it, and unlocks the mutex after each
+ *   method returns.
+ * - thread 1 can share the results of open() with thread 2, and both
+ *   can call extract(), provided they both guard against simultaneous
+ *   use of extract(), and any other methods, with the mutex
  */
 
 #ifndef LIB_MSPACK_H
@@ -124,11 +159,6 @@ extern "C" {
 
 #include <sys/types.h>
 #include <stdlib.h>
-#ifndef _MSC_VER
-# include <unistd.h>
-#endif /* _MSC_VER */
-
-#include <unistd.h>
 
 /**
  * System self-test function, to ensure both library and calling program
@@ -267,7 +297,7 @@ struct mspack_system {
    * @see close(), read(), write(), seek(), tell(), message()
    */
   struct mspack_file * (*open)(struct mspack_system *self,
-			       char *filename,
+			       const char *filename,
 			       int mode);
 
   /**
@@ -362,7 +392,7 @@ struct mspack_system {
    * @see open()
    */
   void (*message)(struct mspack_file *file,
-		  char *format,
+		  const char *format,
 		  ...);
 
   /**
@@ -641,7 +671,7 @@ struct mscabd_cabinet {
    * physical file that the cabinet resides in. This is given by the
    * library user and may be in any format.
    */
-  char *filename;
+  const char *filename;
   
   /** The file offset of cabinet within the physical file it resides in. */
   off_t base_offset;
@@ -901,7 +931,7 @@ struct mscab_decompressor {
    * @see close(), search(), last_error()
    */
   struct mscabd_cabinet * (*open) (struct mscab_decompressor *self,
-				   char *filename);
+				   const char *filename);
 
   /**
    * Closes a previously opened cabinet or cabinet set.
@@ -970,7 +1000,7 @@ struct mscab_decompressor {
    * @see close(), open(), last_error()
    */
   struct mscabd_cabinet * (*search) (struct mscab_decompressor *self,
-				     char *filename);
+				     const char *filename);
 
   /**
    * Appends one mscabd_cabinet to another, forming or extending a cabinet
@@ -1062,7 +1092,7 @@ struct mscab_decompressor {
    */
   int (*extract)(struct mscab_decompressor *self,
 		 struct mscabd_file *file,
-		 char *filename);
+		 const char *filename);
 
   /**
    * Sets a CAB decompression engine parameter.
@@ -1120,7 +1150,7 @@ struct mschmc_file {
 
   /** The filename of the source file that will be added to the CHM. This
    * is passed directly to mspack_system::open(). */
-  char *filename;
+  const char *filename;
 
   /** The full path and filename of the file within the CHM helpfile, a
    * UTF-1 encoded null-terminated string. */
@@ -1182,6 +1212,11 @@ struct mschmd_sec_mscompressed {
 
   /** A pointer to the file which contains the LZX reset table. */
   struct mschmd_file *rtable;
+
+  /** A pointer to the file which contains the LZX span information.
+   * Available only in CHM decoder version 2 and above.
+   */
+  struct mschmd_file *spaninfo;
 };
 
 /**
@@ -1212,7 +1247,7 @@ struct mschmd_header {
    * The filename of the CHM helpfile. This is given by the library user
    * and may be in any format.
    */
-  char *filename;
+  const char *filename;
 
   /** The length of the CHM helpfile, in bytes. */
   off_t length;
@@ -1257,11 +1292,29 @@ struct mschmd_header {
   unsigned int depth;
 
   /**
-   * The number of the root PGMI chunk.
+   * The number of the root PMGI chunk.
    *
    * If there is no index in the CHM helpfile, this will be 0xFFFFFFFF.
    */
   unsigned int index_root;
+
+  /**
+   * The number of the first PMGL chunk. Usually zero.
+   * Available only in CHM decoder version 2 and above.
+   */
+  unsigned int first_pmgl;
+
+  /**
+   * The number of the last PMGL chunk. Usually num_chunks-1.
+   * Available only in CHM decoder version 2 and above.
+   */
+  unsigned int last_pmgl;
+
+  /**
+   * A cache of loaded chunks, filled in by mschm_decoder::fast_find().
+   * Available only in CHM decoder version 2 and above.
+   */
+  unsigned char **chunk_cache;
 };
 
 /**
@@ -1351,7 +1404,7 @@ struct mschm_compressor {
    */
   int (*generate)(struct mschm_compressor *self,
 		  struct mschmc_file file_list[],
-		  char *output_file);
+		  const char *output_file);
 
   /**
    * Specifies whether a temporary file is used during CHM generation.
@@ -1408,7 +1461,7 @@ struct mschm_compressor {
    */
   int (*use_temporary_file)(struct mschm_compressor *self,
 			    int use_temp_file,
-			    char *temp_file);
+			    const char *temp_file);
   /**
    * Sets a CHM compression engine parameter.
    *
@@ -1498,7 +1551,7 @@ struct mschm_decompressor {
    * @see close()
    */
   struct mschmd_header *(*open)(struct mschm_decompressor *self,
-				char *filename);
+				const char *filename);
 
   /**
    * Closes a previously opened CHM helpfile.
@@ -1540,7 +1593,7 @@ struct mschm_decompressor {
    */
   int (*extract)(struct mschm_decompressor *self,
 		 struct mschmd_file *file,
-		 char *filename);
+		 const char *filename);
 
   /**
    * Returns the error code set by the most recently called method.
@@ -1551,7 +1604,7 @@ struct mschm_decompressor {
    * @param  self     a self-referential pointer to the mschm_decompressor
    *                  instance being called
    * @return the most recent error code
-   * @see open(), search()
+   * @see open(), extract()
    */
   int (*last_error)(struct mschm_decompressor *self);
 
@@ -1578,7 +1631,7 @@ struct mschm_decompressor {
    * @see open(), close(), fast_find(), extract()
    */
   struct mschmd_header *(*fast_open)(struct mschm_decompressor *self,
-				     char *filename);
+				     const char *filename);
 
   /**
    * Finds file details quickly.
@@ -1620,7 +1673,7 @@ struct mschm_decompressor {
    */
   int (*fast_find)(struct mschm_decompressor *self,
 		   struct mschmd_header *chm,
-		   char *filename,
+		   const char *filename,
 		   struct mschmd_file *f_ptr,
 		   int f_size);
 };
@@ -1656,12 +1709,21 @@ struct mshlp_decompressor {
 /** msszdd_compressor::set_param() parameter: the missing character */
 #define MSSZDDC_PARAM_MISSINGCHAR (0)
 
+/** msszddd_header::format value - a regular SZDD file */
+#define MSSZDD_FMT_NORMAL (0)
+
+/** msszddd_header::format value - a special QBasic SZDD file */
+#define MSSZDD_FMT_QBASIC (1)
+
 /**
  * A structure which represents an SZDD compressed file.
  *
  * All fields are READ ONLY.
  */
 struct msszddd_header {
+  /** The file format; either #MSSZDD_FMT_NORMAL or #MSSZDD_FMT_QBASIC */
+  int format;
+
   /** The amount of data in the SZDD file once uncompressed. */
   off_t length;
 
@@ -1719,8 +1781,8 @@ struct msszdd_compressor {
    * @see set_param()
    */
   int (*compress)(struct msszdd_compressor *self,
-		  char *input,
-		  char *output,
+		  const char *input,
+		  const char *output,
 		  off_t length);
 
   /**
@@ -1742,7 +1804,7 @@ struct msszdd_compressor {
    * @param  value    the value to set the parameter to
    * @return MSPACK_ERR_OK if all is OK, or MSPACK_ERR_ARGS if there
    *         is a problem with either parameter or value.
-   * @see generate()
+   * @see compress()
    */
   int (*set_param)(struct msszdd_compressor *self,
 		   int param,
@@ -1787,7 +1849,7 @@ struct msszdd_decompressor {
    * @see close()
    */
   struct msszddd_header *(*open)(struct msszdd_decompressor *self,
-				 char *filename);
+				 const char *filename);
 
   /**
    * Closes a previously opened SZDD file.
@@ -1795,9 +1857,7 @@ struct msszdd_decompressor {
    * This closes a SZDD file and frees the msszddd_header associated with
    * it.
    *
-   * The SZDD header pointer is now invalid and cannot be used again. All
-   * msszddd_file pointers referencing that SZDD are also now invalid, and
-   * cannot be used again.
+   * The SZDD header pointer is now invalid and cannot be used again.
    *
    * @param  self     a self-referential pointer to the msszdd_decompressor
    *                  instance being called
@@ -1822,7 +1882,7 @@ struct msszdd_decompressor {
    */
   int (*extract)(struct msszdd_decompressor *self,
 		 struct msszddd_header *szdd,
-		 char *filename);
+		 const char *filename);
 
   /**
    * Decompresses an SZDD file to an output file in one step.
@@ -1842,8 +1902,8 @@ struct msszdd_decompressor {
    * @return an error code, or MSPACK_ERR_OK if successful
    */
   int (*decompress)(struct msszdd_decompressor *self,
-		    char *input,
-		    char *output);
+		    const char *input,
+		    const char *output);
 
   /**
    * Returns the error code set by the most recently called method.
@@ -1854,25 +1914,290 @@ struct msszdd_decompressor {
    * @param  self     a self-referential pointer to the msszdd_decompressor
    *                  instance being called
    * @return the most recent error code
-   * @see open(), search()
+   * @see open(), extract(), decompress()
    */
   int (*last_error)(struct msszdd_decompressor *self);
 };
 
 /* --- support for KWAJ file format ---------------------------------------- */
 
-/** TODO */
-struct mskwaj_compressor {
-  int dummy; 
+/** mskwaj_compressor::set_param() parameter: compression type */
+#define MSKWAJC_PARAM_COMP_TYPE  (0)
+
+/** mskwaj_compressor::set_param() parameter: include the length of the
+ * uncompressed file in the header?
+ */
+#define MSKWAJC_PARAM_INCLUDE_LENGTH (1)
+
+/** KWAJ compression type: no compression. */
+#define MSKWAJ_COMP_NONE (0)
+/** KWAJ compression type: no compression, 0xFF XOR "encryption". */
+#define MSKWAJ_COMP_XOR (1)
+/** KWAJ compression type: LZSS (same method as SZDD) */
+#define MSKWAJ_COMP_SZDD (2)
+/** KWAJ compression type: LZ+Huffman compression */
+#define MSKWAJ_COMP_LZH (3)
+
+/** KWAJ optional header flag: decompressed file length is included */
+#define MSKWAJ_HDR_HASLENGTH (0x01)
+
+/** KWAJ optional header flag: unknown 2-byte structure is included */
+#define MSKWAJ_HDR_HASUNKNOWN1 (0x02)
+
+/** KWAJ optional header flag: unknown multi-sized structure is included */
+#define MSKWAJ_HDR_HASUNKNOWN2 (0x04)
+
+/** KWAJ optional header flag: file name (no extension) is included */
+#define MSKWAJ_HDR_HASFILENAME (0x08)
+
+/** KWAJ optional header flag: file extension is included */
+#define MSKWAJ_HDR_HASFILEEXT (0x10)
+
+/** KWAJ optional header flag: extra text is included */
+#define MSKWAJ_HDR_HASEXTRATEXT (0x20)
+
+/**
+ * A structure which represents an KWAJ compressed file.
+ *
+ * All fields are READ ONLY.
+ */
+struct mskwajd_header {
+  /** The compression type; should be one of #MSKWAJ_COMP_NONE,
+   * #MSKWAJ_COMP_XOR, #MSKWAJ_COMP_SZDD or #MSKWAJ_COMP_LZH
+   */
+  unsigned short comp_type;
+
+  /** The offset in the file where the compressed data stream begins */
+  off_t data_offset;
+
+  /** Flags indicating which optional headers were included. */
+  int headers;
+
+  /** The amount of uncompressed data in the file, or 0 if not present. */
+  off_t length;
+
+  /** output filename, or NULL if not present */
+  char *filename;
+
+  /** extra uncompressed data (usually text) in the header.
+   * This data can contain nulls so use extra_length to get the size.
+   */
+  char *extra;
+
+  /** length of extra uncompressed data in the header */
+  unsigned short extra_length;
 };
 
-/** TODO */
+/**
+ * A compressor for the KWAJ file format.
+ *
+ * All fields are READ ONLY.
+ *
+ * @see mspack_create_kwaj_compressor(), mspack_destroy_kwaj_compressor()
+ */
+struct mskwaj_compressor {
+  /**
+   * Reads an input file and creates a compressed output file in the
+   * KWAJ compressed file format. The KWAJ compression format is quick
+   * but gives poor compression. It is possible for the compressed output
+   * file to be larger than the input file.
+   *
+   * @param  self    a self-referential pointer to the mskwaj_compressor
+   *                 instance being called
+   * @param  input   the name of the file to compressed. This is passed
+   *                 passed directly to mspack_system::open()
+   * @param  output  the name of the file to write compressed data to.
+   *                 This is passed directly to mspack_system::open().
+   * @param  length  the length of the uncompressed file, or -1 to indicate
+   *                 that this should be determined automatically by using
+   *                 mspack_system::seek() on the input file.
+   * @return an error code, or MSPACK_ERR_OK if successful
+   * @see set_param()
+   */
+  int (*compress)(struct mskwaj_compressor *self,
+		  const char *input,
+		  const char *output,
+		  off_t length);
+
+  /**
+   * Sets an KWAJ compression engine parameter.
+   *
+   * The following parameters are defined:
+   *
+   * - #MSKWAJC_PARAM_COMP_TYPE: the compression method to use. Must
+   *   be one of #MSKWAJC_COMP_NONE, #MSKWAJC_COMP_XOR, #MSKWAJ_COMP_SZDD
+   *   or #MSKWAJ_COMP_LZH. The default is #MSKWAJ_COMP_LZH.
+   *
+   * - #MSKWAJC_PARAM_INCLUDE_LENGTH: a boolean; should the compressed
+   *   output file should include the uncompressed length of the input
+   *   file in the header? This adds 4 bytes to the size of the output
+   *   file. A value of zero says "no", non-zero says "yes". The default
+   *   is "no".
+   *
+   * @param  self     a self-referential pointer to the mskwaj_compressor
+   *                  instance being called
+   * @param  param    the parameter to set
+   * @param  value    the value to set the parameter to
+   * @return MSPACK_ERR_OK if all is OK, or MSPACK_ERR_ARGS if there
+   *         is a problem with either parameter or value.
+   * @see generate()
+   */
+  int (*set_param)(struct mskwaj_compressor *self,
+		   int param,
+		   unsigned int value);
+
+
+  /**
+   * Sets the original filename of the file before compression,
+   * which will be stored in the header of the output file.
+   *
+   * The filename should be a null-terminated string, it must be an
+   * MS-DOS "8.3" type filename (up to 8 bytes for the filename, then
+   * optionally a "." and up to 3 bytes for a filename extension).
+   *
+   * If NULL is passed as the filename, no filename is included in the
+   * header. This is the default.
+   *
+   * @param  self     a self-referential pointer to the mskwaj_compressor
+   *                  instance being called
+   * @param  filename the original filename to use
+   * @return MSPACK_ERR_OK if all is OK, or MSPACK_ERR_ARGS if the
+   *         filename is too long
+   */
+  int (*set_filename)(struct mskwaj_compressor *self,
+		      const char *filename);
+
+  /**
+   * Sets arbitrary data that will be stored in the header of the
+   * output file, uncompressed. It can be up to roughly 64 kilobytes,
+   * as the overall size of the header must not exceed 65535 bytes.
+   * The data can contain null bytes if desired.
+   *
+   * If NULL is passed as the data pointer, or zero is passed as the
+   * length, no extra data is included in the header. This is the
+   * default.
+   *
+   * @param  self     a self-referential pointer to the mskwaj_compressor
+   *                  instance being called
+   * @param  data     a pointer to the data to be stored in the header
+   * @param  bytes    the length of the data in bytes
+   * @return MSPACK_ERR_OK if all is OK, or MSPACK_ERR_ARGS extra data
+   *         is too long
+   */
+  int (*set_extra_data)(struct mskwaj_compressor *self,
+			void *data,
+			size_t bytes);
+
+  /**
+   * Returns the error code set by the most recently called method.
+   *
+   * @param  self     a self-referential pointer to the mskwaj_compressor
+   *                  instance being called
+   * @return the most recent error code
+   * @see compress()
+   */
+  int (*last_error)(struct mschm_decompressor *self);
+};
+
+/**
+ * A decompressor for KWAJ compressed files.
+ *
+ * All fields are READ ONLY.
+ *
+ * @see mspack_create_kwaj_decompressor(), mspack_destroy_kwaj_decompressor()
+ */
 struct mskwaj_decompressor {
-  int dummy; 
+  /**
+   * Opens a KWAJ file and reads the header.
+   *
+   * If the file opened is a valid KWAJ file, all headers will be read and
+   * a mskwajd_header structure will be returned.
+   *
+   * In the case of an error occuring, NULL is returned and the error code
+   * is available from last_error().
+   *
+   * The filename pointer should be considered "in use" until close() is
+   * called on the KWAJ file.
+   *
+   * @param  self     a self-referential pointer to the mskwaj_decompressor
+   *                  instance being called
+   * @param  filename the filename of the KWAJ compressed file. This is
+   *                  passed directly to mspack_system::open().
+   * @return a pointer to a mskwajd_header structure, or NULL on failure
+   * @see close()
+   */
+  struct mskwajd_header *(*open)(struct mskwaj_decompressor *self,
+				 const char *filename);
+
+  /**
+   * Closes a previously opened KWAJ file.
+   *
+   * This closes a KWAJ file and frees the mskwajd_header associated
+   * with it. The KWAJ header pointer is now invalid and cannot be
+   * used again.
+   *
+   * @param  self     a self-referential pointer to the mskwaj_decompressor
+   *                  instance being called
+   * @param  kwaj     the KWAJ file to close
+   * @see open()
+   */
+  void (*close)(struct mskwaj_decompressor *self,
+		struct mskwajd_header *kwaj);
+
+  /**
+   * Extracts the compressed data from a KWAJ file.
+   *
+   * This decompresses the compressed KWAJ data stream and writes it to
+   * an output file.
+   *
+   * @param  self     a self-referential pointer to the mskwaj_decompressor
+   *                  instance being called
+   * @param  kwaj     the KWAJ file to extract data from
+   * @param  filename the filename to write the decompressed data to. This
+   *                  is passed directly to mspack_system::open().
+   * @return an error code, or MSPACK_ERR_OK if successful
+   */
+  int (*extract)(struct mskwaj_decompressor *self,
+		 struct mskwajd_header *kwaj,
+		 const char *filename);
+
+  /**
+   * Decompresses an KWAJ file to an output file in one step.
+   *
+   * This opens an KWAJ file as input, reads the header, then decompresses
+   * the compressed data immediately to an output file, finally closing
+   * both the input and output file. It is more convenient to use than
+   * open() then extract() then close(), if you do not need to know the
+   * KWAJ output size or output filename.
+   *
+   * @param  self     a self-referential pointer to the mskwaj_decompressor
+   *                  instance being called
+   * @param  input    the filename of the input KWAJ file. This is passed
+   *                  directly to mspack_system::open().
+   * @param  output   the filename to write the decompressed data to. This
+   *                  is passed directly to mspack_system::open().
+   * @return an error code, or MSPACK_ERR_OK if successful
+   */
+  int (*decompress)(struct mskwaj_decompressor *self,
+		    const char *input,
+		    const char *output);
+
+  /**
+   * Returns the error code set by the most recently called method.
+   *
+   * This is useful for open() which does not return an
+   * error code directly.
+   *
+   * @param  self     a self-referential pointer to the mskwaj_decompressor
+   *                  instance being called
+   * @return the most recent error code
+   * @see open(), search()
+   */
+  int (*last_error)(struct mskwaj_decompressor *self);
 };
 
 #ifdef __cplusplus
-};
+}
 #endif
 
 #endif
