@@ -13,6 +13,10 @@
 #ifndef MSPACK_MSZIP_H
 #define MSPACK_MSZIP_H 1
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 /* MSZIP (deflate) compression / (inflate) decompression definitions */
 
 #define MSZIP_FRAME_SIZE          (32768) /* size of LZ history window */
@@ -110,4 +114,8 @@ extern int mszipd_decompress(struct mszipd_stream *zip, off_t out_bytes);
  */
 void mszipd_free(struct mszipd_stream *zip);
 
+#ifdef __cplusplus
+}
+#endif
+
 #endif
