@@ -1,5 +1,5 @@
-/* cabextract 1.3 - a program to extract Microsoft Cabinet files
- * (C) 2000-2010 Stuart Caie <kyzer@4u.net>
+/* cabextract 1.4 - a program to extract Microsoft Cabinet files
+ * (C) 2000-2011 Stuart Caie <kyzer@4u.net>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -189,7 +189,7 @@ struct cabextract_args args = {
  * cabx_open() when the STDOUT_FNAME pointer is given as a filename, so
  * treat this like a constant rather than a string.
  */
-char *STDOUT_FNAME = "stdout";
+const char *STDOUT_FNAME = "stdout";
 
 /** A special filename. Extracting to this filename will send the output
  * through an MD5 checksum calculator, instead of a file on disk. The
@@ -197,7 +197,7 @@ char *STDOUT_FNAME = "stdout";
  * filename, so treat this like a constant rather than a string. 
  */
 
-char *TEST_FNAME = "test";
+const char *TEST_FNAME = "test";
 
 /** A global MD5 context, used when a file is written to TEST_FNAME */
 struct md5_ctx md5_context;
@@ -223,13 +223,13 @@ static int ensure_filepath(char *path);
 static char *cab_error(struct mscab_decompressor *cd);
 
 static struct mspack_file *cabx_open(struct mspack_system *this,
-                                     char *filename, int mode);
+                                     const char *filename, int mode);
 static void cabx_close(struct mspack_file *file);
 static int cabx_read(struct mspack_file *file, void *buffer, int bytes);
 static int cabx_write(struct mspack_file *file, void *buffer, int bytes);
 static int cabx_seek(struct mspack_file *file, off_t offset, int mode);
 static off_t cabx_tell(struct mspack_file *file);
-static void cabx_msg(struct mspack_file *file, char *format, ...);
+static void cabx_msg(struct mspack_file *file, const char *format, ...);
 static void *cabx_alloc(struct mspack_system *this, size_t bytes);
 static void cabx_free(void *buffer);
 static void cabx_copy(void *src, void *dest, size_t bytes);
@@ -284,7 +284,7 @@ int main(int argc, char *argv[]) {
       "  -s   --single      restrict search to cabs on the command line\n"
       "  -F   --filter      extract only files that match the given pattern\n"
       "  -d   --directory   extract all files to the given directory\n\n"
-      "cabextract %s (C) 2000-2010 Stuart Caie <kyzer@4u.net>\n"
+      "cabextract %s (C) 2000-2011 Stuart Caie <kyzer@4u.net>\n"
       "This is free software with ABSOLUTELY NO WARRANTY.\n",
       VERSION);
     return EXIT_FAILURE;
@@ -520,8 +520,8 @@ static int process_cabinet(char *basename) {
     } /* for (all files in cab) */
 
     /* free the spanning cabinet filenames [not freed by cabd->close()] */
-    for (cab2 = cab->prevcab; cab2; cab2 = cab2->prevcab) free(cab2->filename);
-    for (cab2 = cab->nextcab; cab2; cab2 = cab2->nextcab) free(cab2->filename);
+    for (cab2 = cab->prevcab; cab2; cab2 = cab2->prevcab) free((void*)cab2->filename);
+    for (cab2 = cab->nextcab; cab2; cab2 = cab2->nextcab) free((void*)cab2->filename);
   } /* for (all cabs) */
 
   /* free all loaded cabinets */
@@ -556,8 +556,9 @@ static void load_spanning_cabinets(struct mscabd_cabinet *basecab,
              cab->prevname, cab->previnfo);
     }
     if (!(cab2 = cabd->open(cabd,name)) || cabd->prepend(cabd, cab, cab2)) {
+      fprintf(stderr, "%s: can't prepend %s: %s\n", basename,
+              cab->prevname, cab_error(cabd));
       if (cab2) cabd->close(cabd, cab2);
-      fprintf(stderr, "%s: %s\n", basename, cab_error(cabd));
       break;
     }
     memorise_file(&cab_exts, name, basename);
@@ -575,8 +576,9 @@ static void load_spanning_cabinets(struct mscabd_cabinet *basecab,
              cab->nextname, cab->nextinfo);
     }
     if (!(cab2 = cabd->open(cabd,name)) || cabd->append(cabd, cab, cab2)) {
+      fprintf(stderr, "%s: can't append %s: %s\n", basename,
+              cab->nextname, cab_error(cabd));
       if (cab2) cabd->close(cabd, cab2);
-      fprintf(stderr, "%s: %s\n", basename, cab_error(cabd));
       break;
     }
     memorise_file(&cab_exts, name, basename);
@@ -1070,14 +1072,15 @@ static char *cab_error(struct mscab_decompressor *cd) {
 
 struct mspack_file_p {
   FILE *fh;
-  char *name, regular_file;
+  const char *name;
+  char regular_file;
 };
 
 static struct mspack_file *cabx_open(struct mspack_system *this,
-                                    char *filename, int mode)
+                                    const char *filename, int mode)
 {
   struct mspack_file_p *fh;
-  char *fmode;
+  const char *fmode;
 
   /* Use of the STDOUT_FNAME pointer for a filename means the file should
    * actually be extracted to stdout. Use of the TEST_FNAME pointer for a
@@ -1191,7 +1194,7 @@ static off_t cabx_tell(struct mspack_file *file) {
 #endif
 }
 
-static void cabx_msg(struct mspack_file *file, char *format, ...) {
+static void cabx_msg(struct mspack_file *file, const char *format, ...) {
   va_list ap;
   if (file) {
     fprintf(stderr, "%s: ", ((struct mspack_file_p *) file)->name);
